{-

DEMO-PARSER

-------------------------------------------------------

Vortrag: Monadische Parser
Syntax-Analyse mit Haskell
(Funktionale Programmierung WS 08/09)

Mathias Mierswa (winf2920) und Thomas Stuht (inf6937)

8. Januar 2009

-}

module DemoParser 
where

import Control.Monad
import Text.ParserCombinators.Parsec
import Text.ParserCombinators.Parsec.Expr
import Text.ParserCombinators.Parsec.Token
import Text.ParserCombinators.Parsec.Language

{- Syntaxbaum -}
type Ident = String
data Block = Block [Stmt] deriving Show
data Stmt = DoIt 
            | Assign Ident Expr 
            | If Expr Block Block 
            | While Expr Block
                deriving Show
data Expr = Id Ident 
            | Boolean Bool 
            | UnEx UnOp Expr
            | BinEx BinOp Expr Expr
                deriving Show
data UnOp = Not deriving Show
data BinOp = And | Eq deriving Show

{- Definition der Syntaxbestandteile -}
synDef = emptyDef { commentStart = "{-"
                    , commentEnd = "-}"
                    , nestedComments = True
                    , identStart = letter
                    , identLetter = alphaNum
                    , opStart = oneOf ":!&="
                    , opLetter = oneOf "&="
                    , reservedOpNames= [":=", "!", "&&","=="]
                    , reservedNames = ["TRUE", "FALSE", "DoIt", "IF", "THEN", "ELSE", "ENDIF", "WHILE", "DO", "ENDWHILE"]
                    , caseSensitive = True
                    } 
                    
{- Erzeugung einfacher Parser -}
TokenParser{ parens = m_parens
                    , identifier = m_identifier
                    , reservedOp = m_reservedOp
                    , reserved = m_reserved
                    , semiSep1 = m_semiSep1
                    , whiteSpace = m_whiteSpace
                    } = makeTokenParser synDef
                    
{- Parser fr Expression -}
exprparser :: Parser Expr
exprparser = buildExpressionParser table term <?> "expression"

table = [ [Prefix (m_reservedOp "!" >> return (UnEx Not))]
        , [Infix (m_reservedOp "&&" >> return (BinEx And)) AssocLeft]
        , [Infix (m_reservedOp "==" >> return (BinEx Eq)) AssocLeft]
        ] 
        
term = m_parens exprparser
        <|> liftM Id m_identifier
        <|> (m_reserved "TRUE" >> return (Boolean True))
        <|> (m_reserved "FALSE" >> return (Boolean False))
        
{- Parser fr Statements -}
mainparser :: Parser Block
mainparser = m_whiteSpace >> stmtparser
        where
            stmtparser :: Parser Block
            stmtparser = liftM Block (m_semiSep1 stmt1)
            stmt1 = (m_reserved "DoIt" >> return DoIt)
                    <|> do { i <- m_identifier
                            ; m_reservedOp ":="
                            ; e <- exprparser
                            ; return (Assign i e)
                            } 
                    <|> do { m_reserved "IF"
                            ; c <- exprparser
                            ; m_reserved "THEN"
                            ; p <- stmtparser
                            ; q <- option (Block []) (do{m_reserved "ELSE"; s <- stmtparser; return s})
                            ; m_reserved "ENDIF"
                            ; return (If c p q)
                            }
                    <|> do { m_reserved "WHILE"
                            ; c <- exprparser
                            ; m_reserved "DO"
                            ; p <- stmtparser
                            ; m_reserved "ENDWHILE"
                            ; return (While c p)
                            }
                            
{- Aufruf des Parsers -}
runParsec :: String -> IO ()
runParsec s = case parse mainparser "" s of
                { Left err  -> print err
                ; Right out -> displayAST out -- print out
                }
                
code1 = "IF TRUE DoIt ENDIF"

code2 = "IF   \n TRUE THEN\t\t\tDoIt\n\n ENDIF"

code3 = "DoIt; myVar = string; WHILE DO DoIt ENDWHILE"             
                
code4 = "DoIt; myVar := text42; IF (c == TRUE) THEN DoIt ELSE myVar := andererText99 ENDIF"

code5 = "WHILE !wahr DO DoIt; jahr := jahr2009 IF FALSE THEN jahr := jahr2008 ENDIF ENDWHILE"

code6 = "WHILE !wahr DO DoIt; jahr := jahr2009; IF FALSE THEN jahr := jahr2008 ENDIF ENDWHILE"


parse1 = runParsec code1

parse2 = runParsec code2

parse3 = runParsec code3

parse4 = runParsec code4

parse5 = runParsec code5

parse6 = runParsec code6


{- PrettyPrint-Funktionen -}

mkSpace :: Int -> String
mkSpace i = replicate i ' '

putIndent :: Int -> String -> IO ()
putIndent i s = putStrLn(mkSpace i ++ s)

displayAST :: Block -> IO ()
displayAST b = do {putChar '\n' ; printBlock 0 b }

printBlock :: Int -> Block -> IO ()
printBlock i (Block []) 	= do {putIndent i "+--Block"}
printBlock i (Block stmts) 	= do {putIndent i "+--Block"
								  ; mapM_ (printStmt (i+3)) stmts
								  }

printStmt :: Int -> Stmt -> IO ()
printStmt i (DoIt)          = do {putIndent i "+--DoIt"}
printStmt i (Assign v e)    = do {putIndent i "+--Assign"
								  ; printIdent (i+3) v
								  ; printExpr (i+3) e
								  }								  
printStmt i (If e b1 b2)    = do {putIndent i "+--If"
                                  ; printExpr (i+3) e
                                  ; printBlock (i+3) b1
                                  ; printBlock (i+3) b2
                                  }
printStmt i (While e b)     = do {putIndent i "+--While"
                                  ; printExpr (i+3) e
                                  ; printBlock (i+3) b
                                  }

printExpr :: Int -> Expr -> IO ()
printExpr i (Id ident)      = do {putIndent i "+--Id"
                                  ; printIdent (i+3) ident
                                  }
printExpr i (Boolean b)     = do {putIndent i "+--Boolean"
                                  ; printBool (i+3) b
                                  }
printExpr i (UnEx o e)      = do {putIndent i ("+--UnEx" ++ " " ++ (show o))
                                  ; printExpr (i+3) e
                                  }
printExpr i (BinEx o e1 e2) = do {putIndent i ("+--BinEx" ++ " " ++ (show o))
                                  ; printExpr (i+3) e1
                                  ; printExpr (i+3) e2
                                  }

printIdent :: Int -> Ident -> IO ()
printIdent i ident = do {putIndent i ("+--Ident" ++ " " ++ ident)}

printBool :: Int -> Bool -> IO ()
printBool i b = do {putIndent i ("+--Boolean" ++ " " ++ (show b))}
