{-

Parser-Funktionen in Haskell

-------------------------------------------------------

Vortrag: Monadische Parser
Syntax-Analyse mit Haskell
(Funktionale Programmierung WS 08/09)

Mathias Mierswa (winf2920) und Thomas Stuht (inf6937)

8. Januar 2009

-}

module ParseFct 
where

--import Char

newtype Parser a = MkParser (String -> [(a,String)])

parse :: Parser a -> String -> [(a,String)]
parse (MkParser p) = p

parseOne :: Parser a -> String -> a
parseOne p = fst . head . parse p


instance Monad Parser where
    return a = MkParser (\s -> [(a,s)])
    p >>= f  = MkParser (\s ->
                concat [parse (f a) s' | (a, s') <- parse p s])
                
class Monad m => MonadPlus m where
    mzero :: m a
    mplus :: m a -> m a -> m a
    
instance MonadPlus Parser where
    mzero = MkParser (\s -> [])
    mplus p q = MkParser (\ s -> parse p s ++ parse q s)
    
limit :: Parser a -> Parser a
limit p = MkParser (\s -> case parse p s of
     [] -> []
     (x:xs) -> [x])

{- Parser - Implementierungen -}
item :: Parser Char
item = MkParser (\s -> case s of
     [] -> []
     (x:xs) -> [(x,xs)])
     
sat :: (Char -> Bool) -> Parser Char
sat p = do {x <- item; if p x then return x else mzero}
    
char :: Char -> Parser Char
char c = sat (== c)

string :: String -> Parser String
string [] = return []
string (x:xs) = do {char x; string xs; return (x:xs)}
                  
digit :: Parser Char
digit = sat isDigit

{- module Char -}
isDigit :: Char -> Bool
isDigit c = c >= '0' && c <= '9'

letter :: Parser Char
letter = sat isAlpha

{- module Char -}
isAlpha :: Char -> Bool
isAlpha c = isUpper c || isLower c

alphanum :: Parser Char
alphanum = sat isAlphaNum

{- module Char -}
isAlphaNum :: Char -> Bool
isAlphaNum c = isDigit c || isAlpha c

lower :: Parser Char
lower = sat isLower

{- module Char -}
isLower :: Char -> Bool
isLower c = c >= 'a' && c <= 'z'

upper :: Parser Char
upper = sat isUpper

{- module Char -}
isUpper :: Char -> Bool
isUpper c = c >= 'A' && c <= 'Z'

exOr :: Parser a -> Parser a -> Parser a
exOr p q = MkParser (\s -> case parse p s of
					[] -> parse q s
					ps -> ps)
					
zeroOrMore :: Parser a -> Parser [a]
zeroOrMore p = exOr (oneOrMore p) (return [])

oneOrMore :: Parser a -> Parser [a]
oneOrMore p = do {x <- p; xs <- zeroOrMore p; return (x:xs)}
            
identifier :: Parser String
identifier = do {x <- lower; xs <- zeroOrMore alphanum; return (x:xs)}

nat :: Parser Int
nat = do {xs <- oneOrMore digit; return (read xs)}
		
natural :: Parser Int
natural = token nat

integer :: Parser Int
integer = exOr (do {char '-'; n <- natural; return (-n)}) natural

space :: Parser ()
space = do {zeroOrMore (sat isSpace); return ()}

{- module Char -}
isSpace :: Char -> Bool
isSpace c =  c `elem` " \t\n\r\f\v\xA0"

token :: Parser a -> Parser a
token p = do {space; x <- p; space; return x}

symbol :: String -> Parser String
symbol s = token (string s)



{-- Lazy-Evaluation --}

--zeroOrMore p = exOr (oneOrMore p) (return [])
--oneOrMore p = do {x <- p; xs <- zeroOrMore p; return (x:xs)}
as = zeroOrMore (char 'a')

lazy1 = parse as undefined 
lazy2 = parse as ('a' :undefined) 
lazy3 = parse as ('a' : 'a' :undefined) 
lazy4 = parse as ('a' : 'a' : 'b' :undefined)

force :: Parser a -> Parser a
force p = MkParser (\s -> (fst (head (rs s)), snd (head (rs s))) : tail (rs s))
			where
			rs s = parse p s
        
zeroOrMore' p = force (exOr (oneOrMore' p) (return []))
oneOrMore' p = do { x <- p; xs <- zeroOrMore' p; return (x:xs)}

as' = zeroOrMore' (char 'a')

lazy1' = parse as' undefined 
lazy2' = parse as' ('a' :undefined) 
lazy3' = parse as' ('a' : 'a' :undefined) 
lazy4' = parse as' ('a' : 'a' : 'b' :undefined)


pItem0 = parse item ""
pItem1 = parse item "a"
pItem2 = parse item "hallo"

pChar1 = parse (char 'w') ""
pChar2 = parse (char 'w') "wedel"

pString1 = parse (string "Wedel") "Hamburg"
pString2 = parse (string "Wedel") "Wedel bei Hamburg"

pUpper1 = parse (upper) "wedel"
pUpper2 = parse (upper) "Wedel"

pExOr1 = parse (exOr (string "Wedel") (string "Hamburg")) "Hamburg"
pExOr2 = parse (exOr (string "Wedel") (string "Hamburg")) "Wedel"
pExOr3 = parse (exOr (string "Wedel") (string "Hamburg")) "Pinneberg"

pZeroOrMore1 = parse (zeroOrMore (char 'h')) ""
pZeroOrMore2 = parse (zeroOrMore (char 'h')) "hhhhhsv"
pZeroOrMore3 = parse (zeroOrMore (char 'v')) "hhhhhsv"

pOneOrMore1 = parse (oneOrMore (char 'h')) ""
pOneOrMore2 = parse (oneOrMore (char 'h')) "hhhhhsv"
pOneOrMore3 = parse (oneOrMore (char 'v')) "hhhhhsv"

pIdent0 = parse identifier ""
pIdent1 = parse identifier "123abc"
pIdent2 = parse identifier "ABCDE"
pIdent3 = parse identifier "aBCDE"

pNat0 = parse nat ""
pNat1 = parse nat "1"
pNat2 = parse nat "123"
pNat3 = parse nat "1a"
pNat4 = parse nat "a1"
pNat5 = parse nat "123 abc"

pNatural0 = parse natural ""
pNatural1 = parse natural "1 "
pNatural2 = parse natural " 1"
pNatural3 = parse natural " 1 "
pNatural4 = parse natural " 1 2 4"

pInt0 = parse integer ""
pInt1 = parse integer "a1"
pInt2 = parse integer "1a"
pInt3 = parse integer "1"
pInt4 = parse integer "-1"

pSymbol1 = parse (symbol "Wedel") "\n\t    \tWedel\n\n\t\nbei Hamburg"
pSymbol2 = parse (symbol "Wedel") "\n\t    \tPinneberg\n\n\t\n"
